import asyncio
from typing import Optional

from ftwgl.api import FTWClient
from ftwgl.rcon import RCON


def gameserver_from_dict(server: dict, ftw_client: FTWClient):
    return GameServer(
        id=server['id'],
        dcid=server['DCID'],
        ip=None,
        password=server['config']['password'],
        rcon_password=server['config']['rcon'],
        ftw_client=ftw_client
    )


class GameServer:
    def __init__(self, id, dcid, ip: Optional[str], rcon_password: str, password: str, ftw_client: FTWClient):
        self.id = id
        self.dcid = dcid
        self.ip = ip
        self.rcon_password = rcon_password
        self.password = password
        self._ftw_client = ftw_client
        self._rcon_client: Optional[RCON] = None

    async def wait_until_setup(self):
        if self.ip is not None:
            return

        server = await self._ftw_client.server_get_with_id(self.id)
        while 'ip' not in server['config']:
            await asyncio.sleep(5)
            server = await self._ftw_client.server_get_with_id(self.id)
        self.ip = server['config']['ip']

    def rcon_command(self, command: str) -> str:
        if self._rcon_client is None:
            if self.ip is None:
                raise ValueError(f"GameServer ({self.id}) ip currently is None, likely still booting")
            else:
                self._rcon_client = RCON(server=self.ip, rcon_password=self.rcon_password)

        return self._rcon_client.rcon(command)



import datetime
from typing import List, Optional, Dict, Union

import aiohttp
from aiohttp import ClientResponse

from ftwgl.enum import UserTeamRole, MatchType, GameType


class FTWClient:
    def __init__(self, ftw_api_key: str, ftw_host: str = 'https://ftwgl.net'):
        if ftw_api_key is None or len(ftw_api_key) == 0:
            raise ValueError('ftw_api_key required, but not set')

        self.ftw_host = ftw_host
        self.ftw_api_key = ftw_api_key

    @staticmethod
    async def _handle_response_body(resp: ClientResponse) -> Union[Dict, str]:
        if resp.content_type == 'application/json':
            return await resp.json()
        else:
            return await resp.text()

    async def _handle_response(self, resp: ClientResponse, extract_key: str = None) -> Optional[Union[dict, str, int]]:
        resp_body = await self._handle_response_body(resp)
        if resp.status == 200:
            if extract_key is None:
                return resp_body
            elif extract_key in resp_body:
                return resp_body[extract_key]
            else:
                return None
        else:
            print(f"Request {resp.request_info.method} {resp.request_info.url} failed with {resp.status}")
            print(resp_body)

    async def cup_create(self, name: str, abbreviation: str, playoff_length: int, minimum_roster_size: int,
                         start_date: datetime, roster_lock_date: datetime) -> int:
        request_body = {
            'name': name,
            'abbreviation': abbreviation,
            'playoff_length': playoff_length,
            'minimum_roster_size': minimum_roster_size,
            'start_date': start_date.timestamp(),
            'roster_lock_date': roster_lock_date.timestamp()
        }
        print(request_body)
        async with aiohttp.ClientSession() as session:
            session.headers.add("Authorization", f"{self.ftw_api_key}")
            async with session.post(f"{self.ftw_host}/api/v1/cup", json=request_body) as resp:
                return await self._handle_response(resp, 'cup_id')

    # Creates a user if one does not exist, otherwise updates based on existing discord_id
    async def user_create_or_update(self, discord_id: int, discord_username: str, urt_auth: str):
        request_body = {
            'discord_id': discord_id,
            'discord_username': discord_username,
            'urt_auth': urt_auth
        }
        async with aiohttp.ClientSession() as session:
            session.headers.add("Authorization", f"{self.ftw_api_key}")
            async with session.post(f"{self.ftw_host}/api/v1/user", json=request_body) as resp:
                return await self._handle_response(resp)

    async def team_create(self, creator_discord_id: int, team_name: str, team_tag: str) -> int:
        request_body = {
            'creator_discord_id': creator_discord_id,
            'name': team_name,
            'tag': team_tag
        }
        async with aiohttp.ClientSession() as session:
            session.headers.add("Authorization", f"{self.ftw_api_key}")
            async with session.post(f"{self.ftw_host}/api/v1/team", json=request_body) as resp:
                return await self._handle_response(resp, 'team_id')

    async def cup_add_team(self, team_id: int, cup_id: int):
        request_body = {
            'team_id': team_id,
            'cup_id': cup_id
        }
        async with aiohttp.ClientSession() as session:
            session.headers.add("Authorization", f"{self.ftw_api_key}")
            async with session.post(f"{self.ftw_host}/api/v1/cup/team/add", json=request_body) as resp:
                await self._handle_response(resp)

    async def cup_remove_team(self, team_id: int, cup_id: int):
        request_body = {
            'team_id': team_id,
            'cup_id': cup_id
        }
        async with aiohttp.ClientSession() as session:
            session.headers.add("Authorization", f"{self.ftw_api_key}")
            async with session.post(f"{self.ftw_host}/api/v1/cup/team/remove", json=request_body) as resp:
                await self._handle_response(resp)

    async def cup_set_team_division(self, cup_id: int, team_id: int, division: int):
        request_body = {
            'cup_id': cup_id,
            'team_id': team_id,
            'division': division
        }
        async with aiohttp.ClientSession() as session:
            session.headers.add("Authorization", f"{self.ftw_api_key}")
            async with session.put(f"{self.ftw_host}/api/v1/cup/team/division", json=request_body) as resp:
                await self._handle_response(resp)

    async def team_add_user_or_update_role(self, team_id: int, discord_id: int, role: UserTeamRole):
        request_body = {
            'team_id': team_id,
            'discord_id': discord_id,
            'role': role.value
        }
        async with aiohttp.ClientSession() as session:
            session.headers.add("Authorization", f"{self.ftw_api_key}")
            async with session.post(f"{self.ftw_host}/api/v1/team/user/join", json=request_body) as resp:
                await self._handle_response(resp)

    async def team_remove_user(self, team_id: int, discord_id: int):
        request_body = {
            'team_id': team_id,
            'discord_id': discord_id
        }
        async with aiohttp.ClientSession() as session:
            session.headers.add("Authorization", f"{self.ftw_api_key}")
            async with session.post(f"{self.ftw_host}/api/v1/team/user/quit", json=request_body) as resp:
                await self._handle_response(resp)

    async def match_create(self, cup_id: int, team_ids: List[int], best_of: int,
                           match_type: MatchType, match_date: Optional[datetime.datetime]) -> int:
        request_body = {
            'cup_id': cup_id,
            'team_ids': team_ids,
            'best_of': best_of,
            'match_type': match_type.value,
            'match_date': match_date.timestamp() if match_date is not None else match_date
        }

        async with aiohttp.ClientSession() as session:
            session.headers.add("Authorization", f"{self.ftw_api_key}")
            async with session.post(f"{self.ftw_host}/api/v1/match", json=request_body) as resp:
                return await self._handle_response(resp, 'match_id')

    async def server_locations(self) -> dict:
        async with aiohttp.ClientSession() as session:
            session.headers.add("Authorization", f"{self.ftw_api_key}")
            async with session.get(f"{self.ftw_host}/api/v1/rent/locations") as resp:
                return await self._handle_response(resp)

    async def server_get_with_id(self, server_id: int) -> Optional[dict]:
        async with aiohttp.ClientSession() as session:
            session.headers.add("Authorization", f"{self.ftw_api_key}")
            async with session.get(f"{self.ftw_host}/api/v1/rent/{server_id}") as resp:
                return await self._handle_response(resp)

    async def server_active(self) -> dict:
        async with aiohttp.ClientSession() as session:
            session.headers.add("Authorization", f"{self.ftw_api_key}")
            async with session.get(f"{self.ftw_host}/api/v1/rent/active") as resp:
                return await self._handle_response(resp)

    # Server will likely take a couple of minutes to boot, so use the returned
    async def server_rent(self, match_id: int, dcid: int, gametype: GameType,
                          rcon: str, password: str, ttl_hours: int) -> int:
        request_body = {
            'match_id': match_id,
            'dcid': dcid,
            'gametype': gametype.value,
            'rcon': rcon,
            'password': password,
            'ttl_hours': ttl_hours
        }
        async with aiohttp.ClientSession() as session:
            session.headers.add("Authorization", f"{self.ftw_api_key}")
            async with session.post(f"{self.ftw_host}/api/v1/rent/match", json=request_body) as resp:
                return await self._handle_response(resp, 'id')



from enum import Enum


class UserTeamRole(str, Enum):
    leader = 'leader'
    captain = 'captain'
    member = 'member'
    inactive = 'inactive'
    invited = 'invited'


class MatchType(str, Enum):
    group = 'group'
    quarter_final = 'quarter_final'
    semi_final = 'semi_final'
    silver_final = 'silver_final'
    grand_final = 'grand_final'


class GameType(int, Enum):
    team_survivor = 4
    capture_the_flag = 7



from distutils.core import setup

setup(
    name='ftwgl',
    packages=['ftwgl'],
    version='0.1.11',
    license='MIT',
    description='FTW Python Client',
    author='solitary',
    author_email='solitaryurt@gmail.com',
    url='https://ftwgl.net',
    keywords=['Gaming League', 'FTWGL'],
    install_requires=[
        'aiohttp'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
)



import socket
import time
from threading import Lock


class RCON:
    def __init__(self, server='localhost', rcon_password='password', port=27960, retries=3):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        if ':' in server:
            self.server, self.port = server.split(":")
        else:
            self.server, self.port = server, port

        self.port = int(self.port)
        self.rcon_password = rcon_password
        self.last_cmd = None
        self.retries = retries  # default number of retries
        self.throttle_time = 0.0  # secs to wait between retries
        self.lock = Lock()

        self.socket.connect((self.server, self.port))
        self.socket.settimeout(0.75)

    def _send(self, data):
        ba = bytearray(b'\xFF\xFF\xFF\xFF')
        for c in f'{data}\n'.encode('ascii'):
            ba.append(c)
        self.socket.send(bytes(ba))

    def _recv(self):
        data = None
        try:
            data = self.socket.recv(4096)
        except socket.timeout:
            pass
        except socket.error:
            pass
        return data

    def _cmd(self, cmd):
        self.last_cmd = cmd
        retries = self.retries
        data = None

        while retries > 0 and data is None:
            self._send(cmd)
            data = self._recv()
            if data is None:
                time.sleep(self.throttle_time)
            retries -= 1

        return data[4:] if data is not None else data

    def rcon(self, cmd) -> str:
        """
        Send an rcon command, cmd
        """
        self.lock.acquire()
        reply = self._cmd('rcon "%s" %s' % (self.rcon_password, cmd))
        self.lock.release()
        return reply.decode('ascii', errors='ignore')
